// ◇UTF-8
// ■ libandroid loader
// 2012/01/20 Hiroaki GOTO as GORRY.
// 
// ICONIA TAB A500で「-landroid」を付加して作成したNative Activityが
// 起動できない現象に対策するためのローダ。
// 
// ■ 現象
// ICONIA TAB A500(2.6.36.3+, Acer_A500_7.009.03_AAP_CUS6JP)では、
// NDK作成時に /system/lib/libandroid.so をリンクするために
// 「-landroid」すると、LD_LIBRARY_PATHで先に検索される
// /vendor/lib/libandroid.so を読んでしまう。
// このため、リンクするべき関数がリンクされず、エラーとなる。
// 
// ■ 対策
// 正しい libandroid.so をdlopen()するためのローダを
// 「最初に起動するNative Activity」として最初に起動する。
// このローダがフルパス指定で /system/lib/libandroid.so を読み、
// 続いて「本来起動するべきNative Activity用ライブラリ」を
// 読み込むようにすることで、正しく起動するようになる。


// 実際に起動するNative Activityのライブラリ名と起動関数名
// android:name="android.app.lib_name" で指定するものと同じ。
#define MAIN_SO_NAME "nativetest"
// android:name="android.app.func_name" で指定するものと同じ。
#define MAIN_SO_FUNCNAME "ANativeActivity_onCreate"

// libandroid.soのフルパス
#define LIBANDROID_SO_NAME "/system/lib/libandroid.so"

// =====================================================================

#include <jni.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <dlfcn.h>

#include <android/log.h>
#include <android/native_activity.h>

#ifndef NDEBUG
#define LOGI(...) ((void)__android_log_print(ANDROID_LOG_INFO, "threaded_app", __VA_ARGS__))
#else
#define LOGI(...)
#endif
#define LOGE(...) ((void)__android_log_print(ANDROID_LOG_ERROR, "threaded_app", __VA_ARGS__))

// =====================================================================
// パッケージ名の取得
// =====================================================================

void getJavaPackageName(ANativeActivity* activity, char *buf, int bufsize)
{
	JNIEnv *env = activity->env;

	jclass clazz = (*env)->GetObjectClass(env, activity->clazz);
	jmethodID methodj = (*env)->GetMethodID(env, clazz, "getPackageName", "()Ljava/lang/String;");
	jstring jstr = (jstring)(*env)->CallObjectMethod(env, activity->clazz, methodj);
	if (jstr != NULL) {
		const char* str = (*env)->GetStringUTFChars(env, jstr, NULL);
		strlcpy(buf, str, bufsize);
		(*env)->ReleaseStringUTFChars(env, jstr, str);
	}
}

// =====================================================================
// ローダー本体
// =====================================================================

void AMyLoader(ANativeActivity* activity, void* savedState, size_t savedStateSize) {
	void *handle;
	ANativeActivity_createFunc *func = NULL;
	const char *error;
	char pkgname[1024];
	char path[1024];
	FILE *fin;

	// まず /system/lib/libandroid.so を読み込む
	LOGI("Loading: [%s]", LIBANDROID_SO_NAME);
	handle = dlopen(LIBANDROID_SO_NAME, RTLD_LAZY|RTLD_GLOBAL);
	if (!handle) {
		LOGE("Failed: Loading: [%s]", LIBANDROID_SO_NAME);
		return;
	}

	// 本来起動するべきNative Activityのライブラリファイル名を求める
	getJavaPackageName(activity, pkgname, sizeof(pkgname));
	sprintf(path, "/data/data/%s/lib/lib%s.so", pkgname, MAIN_SO_NAME);

	// ライブラリファイルが存在することを確認する
	LOGI("Check file: [%s]", path);
	fin = fopen(path, "rb");
	if (!fin) {
		LOGE("Not Exist: [%s]", path);
		return;
	}
	fclose(fin);

	// ライブラリファイルを読み込む
	LOGI("Loading: [%s]", path);
	handle = dlopen(path, RTLD_LAZY|RTLD_GLOBAL);
	if (!handle) {
		LOGE("Failed: Loading: [%s]", path);
		error = dlerror();
		if (error)  {
			LOGE("Failed: Loading: [%s]: %s", path, error);
		}
		return;
	}

	// 実行する関数名から実行アドレスを得る
	LOGI("Find function symbol: [%s]", MAIN_SO_FUNCNAME);
	func = dlsym(handle, MAIN_SO_FUNCNAME);
	if (!func) {
		LOGE("Failed: Find function symbol: [%s]", MAIN_SO_FUNCNAME);
		error = dlerror();
		if (error)  {
			LOGE("Failed: Find function symbol: [%s]: %s", MAIN_SO_FUNCNAME, error);
		}
		return;
	}

	// 関数を実行
	LOGI("Call function: [%s]", MAIN_SO_FUNCNAME);
	func(activity, savedState, savedStateSize);

	LOGI("Return from function");
}

// [EOF]
